      module  eigen_libs     
!--   
      use eigen_blacs, only :  eigen_get_blacs_context
      implicit NONE   

      character(32), private ::  CodeNAME = 'EigenExa'
      type  version_t
                           integer ::  Major_Version ! Major version
                           integer ::  Minor_Version ! Minor version
                           integer ::  Patch_Level   ! Patchlevel 0=none
                                                     !,1=a,2=b,...
                           character(32) :: date     ! Release date
                           character(32) :: vcode    ! Version code name
      end type  version_t
      type(version_t), private  :: Eigen_Version ! 2.3L
     &     = version_t (
     &     2, 3, 12,              ! Major, Minor, Patchlevel
     &     'April 12, 2017',      ! Release date
     &     'Sakaki'               ! Version code
     &     )
!--   
      integer, parameter ::  eigen_NB    = 64
!--   
      interface
!--   
      subroutine  eigen_sx( n, nvec, a, lda, w, z, ldz,
     &     m_forward, m_backward, mode )
      integer, intent(IN)    :: n, lda, ldz
                                ! lda and ldz are the leading dimensions
                                ! corresponding to a and z,
                                ! respectively.
                                ! These values must be equal or larger
                                ! than one obtained by calling
                                ! eigen_get_matdims
      integer, intent(IN), optional :: nvec
                                ! the number of eigenvectors computed
                                ! default case, it is set as n
      integer, intent(IN), optional :: m_forward, m_backward
                                ! default values are
                                ! m_forward = 48 and m_backward = 128.
      character, intent(IN), optional :: mode
                                ! = 'A' : eigenvalues and corresponding 
                                !         eigenvectors (default)
                                ! = 'N' : eigenvalues only
                                ! = 'X' : mode 'A' + improve accuracy of
                                !         eigenvalues
      real(8), intent(INOUT) :: a(1:lda, *)
      real(8), intent(OUT)   :: w(*), z(1:ldz, *)
      end subroutine  eigen_sx
!--   
      subroutine  eigen_s( n, nvec, a, lda, w, z, ldz,
     &     m_forward, m_backward, mode )
      integer, intent(IN)    :: n, lda, ldz
                                ! lda and ldz are the leading dimensions
                                ! corresponding to a and z,
                                ! respectively.
                                ! These values must be equal or larger
                                ! than one obtained by calling
                                ! eigen_get_matdims
      integer, intent(IN), optional :: nvec
                                ! the number of eigenvectors computed
                                ! default case, it is set as n
      integer, intent(IN), optional :: m_forward, m_backward
                                ! default values are
                                ! m_forward = 48, and m_backward = 128.
      character, intent(IN), optional :: mode
                                ! = 'A' : eigenvalues and corresponding 
                                !         eigenvectors (default)
                                ! = 'N' : eigenvalues only
                                ! = 'X' : mode 'A' + improve accuracy of
                                !         eigenvalues
      real(8), intent(INOUT) :: a(1:lda, *)
      real(8), intent(OUT)   :: w(*), z(1:ldz, *)
      end subroutine  eigen_s
!--   
      end interface

!     private ::  eigen_loop_calc

      contains
!--   
!--   
      subroutine  eigen_get_version( version, date, vcode )
      use eigen_devel
      implicit NONE
      integer, intent(OUT) :: version
      character*(*), intent(OUT), optional :: date
      character*(*), intent(OUT), optional :: vcode

      version = Eigen_Version%Major_Version * 100
     &     + Eigen_Version%Minor_Version * 10
     &     + Eigen_Version%Patch_Level

      if ( PRESENT(date) ) then
         date = Eigen_Version%date
      end if

      if ( PRESENT(vcode) ) then
         vcode = Eigen_Version%vcode
      end if

      return
      end subroutine  eigen_get_version
!--   
!--   
      subroutine  eigen_show_version( )
      use eigen_devel
      implicit NONE
      character*(256) :: version
      character*(1  ) :: patchlevel
      integer :: i


      i = MIN( 26, Eigen_Version%Patch_Level ) + 1
      patchlevel = ( " abcdefghijklmnopqrstuvwxyz*" (i:i) )

      write( version, '(I1,A,I1,A)' )
     &     Eigen_Version%Major_Version,
     &     '.',Eigen_Version%Minor_Version, trim(patchlevel)

      if ( TRD_inod == 1 ) then
         print*, "## EigenExa version (", trim(version),
     &        ") / (", trim(Eigen_Version%date),
     &        ") / (", trim(Eigen_Version%vcode), ")"
      end if

      return
      end subroutine  eigen_show_version
!--   
!--   
      subroutine  eigen_initialized( flag )
      use eigen_devel
      implicit NONE
!-
      logical, intent(OUT) :: flag

      call eigen_get_initialized( flag )

      return
      end subroutine  eigen_initialized
!--   
!--   
      subroutine  eigen_init( comm, order,
     &     scalapack_context, gridmap )
!-
!$    use OMP_LIB
      use MPI
      use eigen_devel
      use eigen_blacs, only : eigen_blacs_init
      implicit NONE
!-
      integer, intent(IN), optional ::  comm
      character*(*), intent(IN), optional ::  order
                                ! scalapack_context option will be
                                ! supported in future version
      integer, intent(IN), optional ::  scalapack_context
                                ! gridmap option will be supported in a
                                ! future version
      integer, intent(IN), optional ::  gridmap(:,:)
!-
      logical :: flag
      integer ::  n1, n2, n3, i, j, k, ierr
      integer :: topo_type, cart_dim, dims(2), coords(2)
      logical :: periods(2)
      integer :: old_grp, new_grp

      integer, pointer :: tmpgrid(:,:)
      integer          :: comm0
      integer          :: group0, group1, kk0(1), kk1(1)

      integer          :: local_size
!$    integer          :: th0(2), th1(2)

      real(8)          :: s1, s2, d1, d2, buff(4)
      real(8), pointer :: buff1(:), buff2(:)
!     real(8), allocatable :: buff1(:), buff2(:)

      character*1 :: GRID_major = 'C'

!--------------------

      call eigen_timer_reset( 0, 0, 0, 0 )

!--------------------

      call eigen_initialized( flag )

      if ( flag ) then
         if ( TRD_inod == 1 ) then
            print*,"*************"
            print*,"** CAUTION **"
            print*,"*************"
            print*,"You are going to initialize EigenExa,"
            print*,"while EigenExa was not freed at last call."
            print*,"EigenExa restarts again by itself."
         end if
         call eigen_free( )
      end if

!---- Comm setup ---

      call eigen_init_comm_setup( )

!---- OpenMP thread check ---

      call eigen_init_omp_setup( )

!---- Cartesian check ---

      call eigen_init_cartesian_check( )

!---- BLACS setup ---

      call eigen_init_blacs_setup( )

!---- Preliminary Sampling Collective comms. ---

      call eigen_init_collective_comms( )

!---- Final setup ---

      call eigen_set_initialized( )

!--------------------

      return
      contains

!--------------------

      subroutine  eigen_init_comm_setup( )

      if ( PRESENT(comm) ) then
         comm0 = comm
      else
         comm0 = MPI_COMM_WORLD
      end if
      call MPI_Comm_Test_Inter( comm0, flag, ierr )
      if ( flag .OR. ierr /= MPI_SUCCESS ) then
         print*,"*************"
         print*,"** CAUTION **"
         print*,"*************"
         print*,"You are going to initialize EigenExa with"
         print*,"an invalid communicator."
         print*,"EigenExa terminates this run."
         call sleep(1)
         call MPI_Abort( MPI_COMM_WORLD, MPI_ERR_OTHER, ierr )
      endif
      if ( comm0 == MPI_COMM_NULL ) then
         TRD_COMM_WORLD = MPI_COMM_NULL
         x_COMM_WORLD   = MPI_COMM_NULL
         y_COMM_WORLD   = MPI_COMM_NULL
         z_COMM_WORLD   = MPI_COMM_NULL
         w_COMM_WORLD   = MPI_COMM_NULL
         TRD_nnod = 0; TRD_inod = 0
         x_nnod   = 0; x_inod   = 0
         y_nnod   = 0; y_inod   = 0
         z_inod   = 0; z_nnod   = 0
         w_inod   = 0; w_nnod   = 0
      else
         call MPI_Comm_dup( comm0, TRD_COMM_WORLD, ierr )
      end if

      return
      end subroutine  eigen_init_comm_setup

!--------------------

      subroutine  eigen_init_omp_setup( )

!$    if ( TRD_COMM_WORLD /= MPI_COMM_NULL ) then
!$       call MPI_Query_thread( MPI_THREAD_MODE, ierr )
!$       local_size = 1
!$OMP PARALLEL
!$OMP MASTER
!$       local_size = omp_get_num_threads( )
!$OMP END MASTER
!$OMP END PARALLEL
!$       th0(1) = local_size
!$       th0(2) = -th0(1)
!$       call MPI_Allreduce( th0, th1, 2, MPI_INTEGER, MPI_MAX,
!$   &        TRD_COMM_WORLD, ierr )
!$       j = th1(1) + th1(2)
!$       if ( j /= 0 ) then
!$          call MPI_Barrier( TRD_COMM_WORLD, ierr )
!$          call flush( 6 )
!$          if ( TRD_inod == 1 ) then
!$             print*,"*************"
!$             print*,"** CAUTION **"
!$             print*,"*************"
!$             print*,"EigenExa supports only homogeneous thread"
!$   &              // " setting !"
!$             print*,"EigenExa terminates this run."
!$             print*,"*************"
!$          end if
!$          do j=1,2
!$             call MPI_Barrier( TRD_COMM_WORLD, ierr )
!$             call flush( 6 )
!$          end do
!$          call MPI_Barrier( TRD_COMM_WORLD, ierr )
!$          call sleep(1)
!$          call MPI_Abort( MPI_COMM_WORLD, MPI_ERR_OTHER, ierr )
!$       end if
!$    end if

      return
      end subroutine  eigen_init_omp_setup

!--------------------

      subroutine  eigen_init_cartesian_check( )

      if ( TRD_COMM_WORLD /= MPI_COMM_NULL ) then

!---- Cartesian check ---

         call MPI_Comm_size( TRD_COMM_WORLD, TRD_nnod, ierr )
         call MPI_Comm_rank( TRD_COMM_WORLD, TRD_inod, ierr )
         TRD_inod = TRD_inod+1

         call MPI_Topo_test( TRD_COMM_WORLD, topo_type, ierr )
         if ( topo_type == MPI_CART ) then
            call MPI_Cartdim_get( TRD_COMM_WORLD, cart_dim, ierr )
         else
            cart_dim = 1
         end if

!---- 2D Cartesian setup ---

         if ( cart_dim == 2 ) then

            call MPI_Cart_get( TRD_COMM_WORLD, cart_dim,
     &           dims, periods, coords, ierr )
            x_nnod = dims(1)
            y_nnod = dims(2)
            x_inod = coords(1) +1 ! fixed
            y_inod = coords(2) +1 ! fixed

            GRID_major = 'R'
            if ( PRESENT(order) ) then
               if ( x_inod == 1 .AND. y_inod == 1 ) then
                  if ( order(1:1) == 'C' .or. order(1:1) == 'c' )
     &                 then
                     print*,"*************"
                     print*,"** CAUTION **"
                     print*,"*************"
                     print*,"The MPI_CART you specified is based on"
     &                    // " R-major,"
                     print*,"but you also specified C-major option."
                     print*,"EigenExa solve this conflict by taking"
     &                    // " R-major,"
                  end if
               end if
            end if

         else                   ! if ( cart_dim == 2

!---- Setup 2D process map ---

            x_nnod = INT(SQRT(DBLE(TRD_nnod)))
            i = 1               ! minimum factor, x_nnod must be
!     multiple of k
            if ( MOD(TRD_nnod, i) == 0 ) then
               k = i
            else
               k = 1
            end if
            do
               if ( x_nnod <= k ) exit
               if ( MOD(x_nnod, k) == 0 .AND.
     &              MOD(TRD_nnod, x_nnod) == 0 ) exit
               x_nnod = x_nnod-1
            end do              !!
            y_nnod = TRD_nnod/x_nnod

            if ( PRESENT(order) ) then
               GRID_major = order(1:1)
            else
               GRID_major = 'C'
            end if

            if ( GRID_major == 'R' .or. GRID_major == 'r' ) then
               GRID_major = 'R'
            else
               GRID_major = 'C'
            endif

            call eigen_set_grid_major( GRID_major )

            if ( GRID_major == 'R' ) then
!     row-major
               x_inod =    (TRD_inod-1)/y_nnod +1
               y_inod = MOD(TRD_inod-1, y_nnod)+1
            else
!     column-major
!     ** EigenK adopts column-major in default
!     The process ordering on the Earth Simulator was done in the way
!     such as incrmental from internal processors to external nodes.
!     We want to make the processes, which possess a distributed vector,
!     close physically on the network connection.
!     These are the historical reason to adopt column-major.
               x_inod = MOD(TRD_inod-1, x_nnod)+1
               y_inod =    (TRD_inod-1)/x_nnod +1
            end if

         end if                 ! if ( cart_dim == 2 ) then

!---- x-Comm & y-Comm setup ---

         call MPI_Comm_split( TRD_COMM_WORLD, y_inod, x_inod,
     &        x_COMM_WORLD, ierr )
         call MPI_Comm_split( TRD_COMM_WORLD, x_inod, y_inod,
     &        y_COMM_WORLD, ierr )

!---- Parameter setup for datacast_dbl (row-column vector redistribution) ---

         n1 = MAX(x_nnod, y_nnod)
         n2 = MIN(x_nnod, y_nnod)
         do
            if ( n1 == n2 ) then
               n_common = n1
               exit
            end if
            n3 = n1-n2
            n1 = MAX(n2, n3)
            n2 = MIN(n2, n3)
         end do                 !!

         if ( ASSOCIATED( p0_ ) ) then
            deallocate( p0_ )
         end if
         allocate ( p0_(1:MAX(x_nnod, y_nnod)) )

         if ( ASSOCIATED( q0_ ) ) then
            deallocate( q0_ )
         end if
         allocate ( q0_(1:MAX(x_nnod, y_nnod)) )

         p0_(:) = -1
         q0_(:) = -1
         do i = 1, x_nnod
            if ( MOD(i-1, n_common) == MOD(y_inod-1, n_common) ) then
               n1 = y_inod-i
               if ( n1 >= 0 ) then
                  do j = 1, x_nnod
                     k = +n1+(j-1)*y_nnod
                     if ( MOD(k, x_nnod) == 0 ) then
                        p0_(i) = k/x_nnod
                        q0_(i) = (j-1)
                        exit
                     end if
                  end do        ! j
               else
                  do j = 1, y_nnod
                     k = -n1+(j-1)*x_nnod
                     if ( MOD(k, y_nnod) == 0 ) then
                        q0_(i) = k/y_nnod
                        p0_(i) = (j-1)
                        exit
                     end if
                  end do        ! j
               end if
            end if
         end do                 ! i
         p0_(:) = p0_(:)+1
         q0_(:) = q0_(:)+1


         diag_0 = 0
         diag_1 = 0
         do i = 1, y_nnod/n_common
            j = (i-1)*y_nnod+y_inod
            k = MOD(j-1, x_nnod)+1
            if ( k == x_inod ) then
               diag_0 = i
               diag_1 = (j-1)/x_nnod+1
               exit
            end if
         end do                 ! i_1

!---- extra-Comm setup for datacast_dbl ---

         if ( n_common > 1 ) then
            w_inod =     (x_inod-1)/ n_common +1
            z_inod = MOD((x_inod-1), n_common)+1
            w_nnod = x_nnod/n_common
            z_nnod = n_common
            call MPI_Comm_split( x_COMM_WORLD, w_inod-1, z_inod-1,
     &           z_COMM_WORLD, ierr )
            call MPI_Comm_split( x_COMM_WORLD, z_inod-1, w_inod-1,
     &           w_COMM_WORLD, ierr )
         else
            z_COMM_WORLD = MPI_COMM_SELF
            w_COMM_WORLD = x_COMM_WORLD
            z_inod = 1; z_nnod = 1
            w_inod = x_inod; w_nnod = x_nnod
         end if

!---- Exceptional setup ---

      else                      !  if ( TRD_COMM_WORLD /= MPI_COMM_NULL

         x_COMM_WORLD = MPI_COMM_NULL
         y_COMM_WORLD = MPI_COMM_NULL
         z_COMM_WORLD = MPI_COMM_NULL
         w_COMM_WORLD = MPI_COMM_NULL
         x_nnod = 0; x_inod = 0
         y_nnod = 0; y_inod = 0
         z_inod = 0; z_nnod = 0
         w_inod = 0; w_nnod = 0

         GRID_major = 'C'
         call eigen_set_grid_major( GRID_major )

      end if                    !  if ( TRD_COMM_WORLD /= MPI_COMM_NULL

      return
      end subroutine  eigen_init_cartesian_check

!--------------------

      subroutine  eigen_init_blacs_setup( )

#if TIMER_PRINT>1
      if ( TRD_inod <= 1 ) then
         print*,"GRID major ",GRID_major," is specified."
      end if
#endif

      if ( TRD_COMM_WORLD == MPI_COMM_NULL ) then
         comm0  = MPI_COMM_SELF
         x_nnod = 1; x_inod = 1
         y_nnod = 1; y_inod = 1
      else
         comm0  = TRD_COMM_WORLD
      end if

      call Eigen_BLACS_Init( comm0, x_nnod, y_nnod, GRID_Major )

      if ( TRD_COMM_WORLD == MPI_COMM_NULL ) then
         x_nnod = 0; x_inod = 0
         y_nnod = 0; y_inod = 0
      end if

      return
      end subroutine  eigen_init_blacs_setup

!--------------------

      subroutine  eigen_init_collective_comms( )

      if ( TRD_COMM_WORLD /= MPI_COMM_NULL ) then

!---- Measure the overhead of collective communications ---

         call MPI_Barrier(TRD_COMM_WORLD,ierr)
         call MPI_Barrier(TRD_COMM_WORLD,ierr)

!$OMP PARALLEL
!$OMP+          PRIVATE(i,j,d1,d2,buff1,buff2)

         allocate( buff1(2048), buff2(2048) )

!$OMP MASTER
         s1 = ZERO
         s2 = ZERO
!$OMP END MASTER
!$OMP BARRIER

         do  i = 1, 10
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call MPI_Barrier(x_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s1 = s1 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

!$OMP MASTER
         call MPI_Barrier(TRD_COMM_WORLD,ierr)
         call MPI_Barrier(TRD_COMM_WORLD,ierr)
!$OMP END MASTER
!$OMP BARRIER

         do  i = 1, 10
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call MPI_Barrier(y_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s2 = s2 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

!$OMP MASTER
         call MPI_Barrier(TRD_COMM_WORLD,ierr)
         call MPI_Barrier(TRD_COMM_WORLD,ierr)

         buff(1) = s1 / 5
         buff(2) = s2 / 5
         call MPI_Allreduce( buff(1), buff(3), 2, MPI_DOUBLE_PRECISION,
     &        MPI_SUM, TRD_COMM_WORLD, ierr )
         Barrier_Overhead_x = buff(3) / TRD_nnod
         Barrier_Overhead_y = buff(4) / TRD_nnod
!$OMP END MASTER

!--------------------

!$OMP MASTER
         s1 = ZERO
         s2 = ZERO
!$OMP END MASTER
!$OMP BARRIER

         do  i = 1, 10
!$OMP MASTER
            call reduce_dbl(buff1(1),buff2(1),1024,0,y_COMM_WORLD)
!$OMP END MASTER
!$OMP BARRIER
            continue
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call reduce_dbl(buff1(1),buff2(1),2048,0,x_COMM_WORLD)
            call MPI_Barrier(x_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s1 = s1 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

         do  i = 1, 10
!$OMP MASTER
            call reduce_dbl(buff1(1),buff2(1),1024,0,x_COMM_WORLD)
!$OMP END MASTER
!$OMP BARRIER
            continue
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call reduce_dbl(buff1(1),buff2(1),2048,0,y_COMM_WORLD)
            call MPI_Barrier(y_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s2 = s2 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

!$OMP MASTER
         buff(1) = s1 / 5
         buff(2) = s2 / 5
         call MPI_Allreduce( buff(1), buff(3), 2, MPI_DOUBLE_PRECISION,
     &        MPI_SUM, TRD_COMM_WORLD, ierr )
         Reduce_Overhead_x = buff(3) / TRD_nnod - Barrier_Overhead_x
         Reduce_Overhead_y = buff(4) / TRD_nnod - Barrier_Overhead_y
!$OMP END MASTER

!--------------------

!$OMP MASTER
         s1 = ZERO
         s2 = ZERO
!$OMP END MASTER
!$OMP BARRIER

         do  i = 1, 10
!$OMP MASTER
            call reduce_dbl(buff1(1),buff2(1),1024,0,y_COMM_WORLD)
!$OMP END MASTER
!$OMP BARRIER
            continue
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call reduce_dbl(buff1(1),buff2(1),1024,0,x_COMM_WORLD)
!$OMP END MASTER
            continue
!$OMP MASTER
            call reduce_dbl(buff2(1),buff1(1),1024,0,x_COMM_WORLD)
            call MPI_Barrier(x_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s1 = s1 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

         do  i = 1, 10
!$OMP MASTER
            call reduce_dbl(buff1(1),buff2(1),1024,0,x_COMM_WORLD)
!$OMP END MASTER
!$OMP BARRIER
            continue
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call reduce_dbl(buff1(1),buff2(1),1024,0,y_COMM_WORLD)
!$OMP END MASTER
            continue
!$OMP MASTER
            call reduce_dbl(buff2(1),buff1(1),1024,0,y_COMM_WORLD)
            call MPI_Barrier(y_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s2 = s2 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

!$OMP MASTER
         buff(1) = s1 / 5
         buff(2) = s2 / 5
         call MPI_Allreduce( buff(1), buff(3), 2, MPI_DOUBLE_PRECISION,
     &        MPI_SUM, TRD_COMM_WORLD, ierr )
         Reduce_cont_Overhead_x = buff(3) / TRD_nnod
     &        - Barrier_Overhead_x - Reduce_Overhead_x
         Reduce_cont_Overhead_y = buff(4) / TRD_nnod
     &        - Barrier_Overhead_y - Reduce_Overhead_y
!$OMP END MASTER

!--------------------

!$OMP MASTER
         s1 = ZERO
         s2 = ZERO
!$OMP END MASTER
!$OMP BARRIER

         do  i = 1, 10
!$OMP MASTER
            call bcast_dbl(buff1(1),1024,1,0,y_COMM_WORLD)
!$OMP END MASTER
!$OMP BARRIER
            continue
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call bcast_dbl(buff1(1),2048,1,0,x_COMM_WORLD)
            call MPI_Barrier(x_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s1 = s1 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

         do  i = 1, 10
!$OMP MASTER
            call bcast_dbl(buff1(1),1024,1,0,x_COMM_WORLD)
!$OMP END MASTER
!$OMP BARRIER
            continue
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call bcast_dbl(buff1(1),2048,1,0,y_COMM_WORLD)
            call MPI_Barrier(y_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s2 = s2 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

!$OMP MASTER
         buff(1) = s1 / 5
         buff(2) = s2 / 5
         call MPI_Allreduce( buff(1), buff(3), 2, MPI_DOUBLE_PRECISION,
     &        MPI_SUM, TRD_COMM_WORLD, ierr )
         Bcast_Overhead_x = buff(3) / TRD_nnod - Barrier_Overhead_x
         Bcast_Overhead_y = buff(4) / TRD_nnod - Barrier_Overhead_y
!$OMP END MASTER

!--------------------

!$OMP MASTER
         s1 = ZERO
         s2 = ZERO
!$OMP END MASTER
!$OMP BARRIER

         do  i = 1, 10
!$OMP MASTER
            call bcast_dbl(buff1(1),1024,1,0,y_COMM_WORLD)
!$OMP END MASTER
!$OMP BARRIER
            continue
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call bcast_dbl(buff1(1),1024,1,0,x_COMM_WORLD)
!$OMP END MASTER
            continue
!$OMP MASTER
            call bcast_dbl(buff2(1),1024,1,0,x_COMM_WORLD)
            call MPI_Barrier(x_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s1 = s1 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

         do  i = 1, 10
!$OMP MASTER
            call bcast_dbl(buff1(1),1024,1,0,x_COMM_WORLD)
!$OMP END MASTER
!$OMP BARRIER
            continue
!$OMP MASTER
            call MPI_Barrier(TRD_COMM_WORLD,ierr)
            d1 = eigen_get_wtime()
            call bcast_dbl(buff1(1),1024,1,0,y_COMM_WORLD)
!$OMP END MASTER
            continue
!$OMP MASTER
            call bcast_dbl(buff2(1),1024,1,0,y_COMM_WORLD)
            call MPI_Barrier(y_COMM_WORLD,ierr)
            d2 = eigen_get_wtime()
            if ( i > 5 ) s2 = s2 + (d2-d1)
!$OMP END MASTER
!$OMP BARRIER
         end do

!$OMP MASTER
         buff(1) = s1 / 5
         buff(2) = s2 / 5
         call MPI_Allreduce( buff(1), buff(3), 2, MPI_DOUBLE_PRECISION,
     &        MPI_SUM, TRD_COMM_WORLD, ierr )
         Bcast_cont_Overhead_x = buff(3) / TRD_nnod
     &        - Barrier_Overhead_x - Bcast_Overhead_x
         Bcast_cont_Overhead_y = buff(4) / TRD_nnod
     &        - Barrier_Overhead_y - Bcast_Overhead_y
!$OMP END MASTER

         deallocate( buff1, buff2 )

!$OMP END PARALLEL

!--------------------

         if ( Reduce_cont_Overhead_x < ZERO ) then
            Reduce_cont_Overhead_x = ZERO
         end if
         if ( Reduce_cont_Overhead_y < ZERO ) then
            Reduce_cont_Overhead_y = ZERO
         end if
         if ( Bcast_cont_Overhead_x  < ZERO ) then
            Bcast_cont_Overhead_x  = ZERO
         end if
         if ( Bcast_cont_Overhead_y  < ZERO ) then
            Bcast_cont_Overhead_y  = ZERO
         end if

#if TIMER_PRINT>0
         if ( TRD_inod == 1 ) then
            print*,";----- MPI in-(x/y)-group overhead -----"
            print*,"Overhead Barrier(x) :: ", Barrier_Overhead_x
            print*,"Overhead Barrier(y) :: ", Barrier_Overhead_y
            print*,"Overhead Bcast(x)   :: ", Bcast_Overhead_x
            print*,"Overhead Bcast(y)   :: ", Bcast_Overhead_y
            print*,"Overhead Reduce(x)  :: ", Reduce_Overhead_x
            print*,"Overhead Reduce(y)  :: ", Reduce_Overhead_y
            print*,"Overhead Bcast_cont(x)   :: ",
     &           Bcast_cont_Overhead_x
            print*,"Overhead Bcast_cont(y)   :: ",
     &           Bcast_cont_Overhead_y
            print*,"Overhead Reduce_cont(x)  :: ",
     &           Reduce_cont_Overhead_x
            print*,"Overhead Reduce_cont(y)  :: ",
     &           Reduce_cont_Overhead_y
         end if
#endif

!--------------------

      end if

      end subroutine  eigen_init_collective_comms

      end subroutine  eigen_init
!--   
!--   
      subroutine  eigen_free( flag )
      use MPI
      use eigen_devel
      use eigen_blacs, only : eigen_blacs_exit
      implicit NONE
      integer, intent(IN), optional ::  flag
!-
      integer                ::  ierr
      logical                ::  local_flag
      real(8)                ::  tmp


      call eigen_get_initialized( local_flag )

      if ( .NOT. local_flag ) then
         return
      end if

      if ( TRD_COMM_WORLD /= MPI_COMM_NULL ) then

         call MPI_Comm_free( x_COMM_WORLD, ierr )
         call MPI_Comm_free( y_COMM_WORLD, ierr )

         if ( n_common > 1 ) then
            call MPI_Comm_free( z_COMM_WORLD, ierr )
            call MPI_Comm_free( w_COMM_WORLD, ierr )
         end if

         call MPI_Comm_free( TRD_COMM_WORLD, ierr )

         deallocate ( p0_)
         deallocate ( q0_)

         if ( PRESENT(flag) ) then
            if ( flag == 1 ) then
               tmp = eigen_timer_print( 'EigenExa(finalized)' )
            end if
         end if
         call eigen_timer_reset( 0, 0, 0, 0 )

      end if

!     call Eigen_BLACS_Exit( )

      TRD_COMM_WORLD = MPI_COMM_WORLD

      call eigen_unset_initialized( )


      return
      end subroutine  eigen_free
!--   
!--   
      subroutine  eigen_get_matdims( n, nx, ny )
      use eigen_devel
      implicit NONE
      integer, intent(IN)    :: n
      integer, intent(OUT)   :: nx, ny
!-
      integer :: NPROW, NPCOL, NB
      integer :: n1, nm, nmz, nmw, nn, larray

      integer :: lddw, lddz, nxx

      integer(8) :: LX1, LX2


      if ( n <= 0 ) then
         nx = -1; ny = -1
         return
      endif

      NPROW = x_nnod
      NPCOL = y_nnod

      n1 = ((n-1)/NPROW+1)
      call CSTAB_get_optdim( n1, 6, 16*4, 16*4*2, nm )

      NB  = eigen_NB

      nmz = ((n-1)/NPROW+1)
      nmz = ((nmz-1)/NB+1)*NB+1
      nn  = nmz
      nmz = (n-1)/NB+1
      nmz = ((nmz-1)/NPROW+1)*NB
!     Fix on version 2.2b
!     to avoid unexpected SIGSEGV,
!     use the maximum of nn and nmz.
      nmz = MAX(nn, nmz)

      nmw = ((n-1)/NPCOL+1)
      nmw = ((nmw-1)/NB+1)*NB+1
      nn  = nmw
      nmw = (n-1)/NB+1
      nmw = ((nmw-1)/NPCOL+1)*NB
!     Fix on version 2.2b
!     to avoid unexpected SIGSEGV,
!     use the maximum of nn and nmw.
      nmw = MAX(nn, nmw)


      larray = MAX(nmz, nm)*nmw

      nx = nm
      ny = (larray-1)/nm+1

#if 0
!     
!     check whether intermediate datasize is valid with in integer(4).
!     
      NB = eigen_NB
      NB = MIN(NB, N)

      lddz = (n-1)/NPROW+1
      lddz = ((lddz-1)/NB+1)*NB
      lddw = (n-1)/NPCOL+1
      lddw = ((lddw-1)/NB+1)*NB
      nxx  = (N-1)/NPCOL+1

      LX1 = lddw; LX1 = LX1 * lddw
      LX2 = nxx;  LX2 = LX2 * n

      if ( LX1 >= (2_8)**31 .OR. LX2 >= (2_8)**31 ) then
         print*, "Warning :: oversized problem !!"
         larray = -1
         nx     = -1
         ny     = -1
      endif
#endif

      return
      end subroutine  eigen_get_matdims
!--   
!--   
!     ******* ! important ! API has been changed ! *********************
!     Since the size of intermediate data may overfolow integer(4),
!     version 2.3b has changed the datatype of the function  result
!     from integer to integer(8).
!     ******************************************************************
      integer(8) function  eigen_memory_internal( n, lda, ldz,
     &     m1_opt, m0_opt )
     &     result( byte )
!$    use OMP_LIB
      use MPI
      use eigen_devel
      implicit NONE
      integer, intent(IN)    :: n, lda, ldz
      integer, intent(IN), optional :: m1_opt, m0_opt
!-
      integer :: m1, m0
      integer :: MYROW, MYCOL, NPROW, NPCOL, NP, NQ, NB
      integer :: local_rank, local_size
      integer :: m , nv, nm, ierr

      integer :: l_uv_t
      integer :: lddz, lddw, LWORK, LIWORK
      integer :: na, nx

      integer, parameter     :: nm_max_L1 = 16*4
      integer, parameter     :: nm_max_L2 = 16*6

      integer(8) :: nax(2)
      integer(8) :: byte_tridi
      integer(8) :: byte_dcx
      integer(8) :: byte_trbak

      integer, external :: NUMROC

      integer, parameter :: REAL_SIZE = 8_8 ! must be in integer(8)
      integer, parameter :: INT_SIZE  = 4_8 ! must be in integer(8)

      include 'CSTAB.h'


      if ( n <= 0 ) then
         byte = -1
         return
      endif

      local_rank = 0
      local_size = 1
!$    local_rank = omp_get_thread_num()
!$    local_size = omp_get_num_threads()

      if ( PRESENT(m1_opt) ) then
         m1 = m1_opt
      else
         m1 = 48
      end if

      if ( PRESENT(m0_opt) ) then
         m0 = m0_opt
      else
         m0 = 128
      end if

!     ===== for TRIDI ====

      nx = (n-1)/x_nnod+1 +2
      call CSTAB_get_optdim(nx, 6, nm_max_L1, nm_max_L2, nv)

      l_uv_t = MAX(2*nv,
     &     2*(n+MAX(3,y_nnod/n_common)),
     &     2*8*m1,
     &     nv+3+4*m1)

      LWORK = MAX(3*m1, (nm*m1+n_columns))
     &     + 2*(l_uv_t+n_columns)
     &     + (nv+n_columns)
     &     + 2*(nv*2*m1+2*n_columns)
     &     + 4*(nv*local_size+n_columns)

      byte_tridi = LWORK * REAL_SIZE

!     ===== for DCX ====

      NB = eigen_NB
      NB = MIN(NB, N)

      NPROW = x_nnod
      NPCOL = y_nnod
      MYROW = x_inod-1
      MYCOL = y_inod-1
      NP = NUMROC( n, NB, MYROW, 0, NPROW )
      NQ = NUMROC( n, NB, MYCOL, 0, NPCOL )

      lddz = (n-1)/NPROW+1
      lddz = ((lddz-1)/NB+1)*NB
      lddw = (n-1)/NPCOL+1
      lddw = ((lddw-1)/NB+1)*NB

      nx     = (N-1)/NPCOL+1
      LWORK  = MAX(1+6*N+2*NP*(NQ+MAX(NQ,NB)), lddz*lddw, ldz*nx)
     &     + n
      LIWORK = 2+7*n+8*NPCOL

      byte_dcx = LWORK * REAL_SIZE + LIWORK * INT_SIZE

!     ===== for TRBAK ====

      m    = MIN(nsm, m0)
      if ( m < 1 ) m = 1

      na   = (n-1)/y_nnod+1
      na   = na  +MOD(na  -1,2)
      call CSTAB_get_optdim(lda, 9, 16*4, 16*6, nm)

      LWORK = n
     &     + 3*(MAX(nm*m,n)+n_columns)
     &     + 4*(na*m+ns0+n_columns)
     &     + m*nm
      LIWORK = MAX(((m-1)/y_nnod+1)*((n-1)/x_nnod+1),n)

      byte_trbak = LWORK * REAL_SIZE + LIWORK * INT_SIZE


      nax(1) = MAX(byte_tridi, byte_dcx, byte_trbak)
      call MPI_Allreduce( nax(1), nax(2), 1, MPI_INTEGER8,
     &     MPI_MAX, TRD_COMM_WORLD, ierr )
      byte = nax(2)


      return
      end function  eigen_memory_internal
!--   
!--   
      subroutine  eigen_get_procs( procs, x_procs, y_procs )
      use eigen_devel
      implicit NONE
      integer, intent(OUT)   ::  procs, x_procs, y_procs


      procs   = TRD_nnod
      x_procs = x_nnod
      y_procs = y_nnod


      return
      end subroutine  eigen_get_procs
!--   
!--   
      subroutine  eigen_get_id( id, x_id, y_id )
      use eigen_devel
      implicit NONE
      integer, intent(OUT)   ::  id, x_id, y_id


      id   = TRD_inod
      x_id = x_inod
      y_id = y_inod


      return
      end subroutine  eigen_get_id
!--   
!--   
      subroutine  eigen_get_comm( comm, x_comm, y_comm )
      use eigen_devel
      implicit NONE
      integer, intent(OUT)   ::  comm, x_comm, y_comm


      comm   = TRD_COMM_WORLD
      x_comm = x_COMM_WORLD
      y_comm = y_COMM_WORLD


      return
      end subroutine  eigen_get_comm
!--   
!--   
      integer function  eigen_loop_start( istart, nnod, inod )
     &     result( ret )
      implicit NONE
      integer, intent(IN)    ::  istart, nnod, inod

      ret = (istart + nnod - 1 - inod) / nnod + 1

      return
      end function  eigen_loop_start
!--   
!--   
      integer function  eigen_loop_end( iend, nnod, inod )
     &     result( ret )
      implicit NONE
      integer, intent(IN)    ::  iend, nnod, inod

      ret = (iend + nnod - 0 - inod) / nnod + 0

      return
      end function  eigen_loop_end
!--   
!--   
      integer function  eigen_translate_l2g( ictr, nnod, inod )
     &     result( ret )
      implicit NONE
      integer, intent(IN)    ::  ictr, nnod, inod

      ret = (ictr-1)*nnod+inod

      return
      end function  eigen_translate_l2g
!--   
!--   
      integer function  eigen_translate_g2l( ictr, nnod, inod )
     &     result( ret )
      implicit NONE
      integer, intent(IN)    ::  ictr, nnod, inod

      ret = (ictr-1)/nnod+1

      return
      end function  eigen_translate_g2l
!--   
!--   
      integer function  eigen_owner_node( ictr, nnod, inod )
     &     result( ret )
      implicit NONE
      integer, intent(IN)    ::  ictr, nnod, inod

      ret = MOD(ictr-1, nnod)+1

      return
      end function  eigen_owner_node
!--   
!--   
      integer function  eigen_owner_index( ictr, nnod, inod )
     &     result( ret )
      implicit NONE
      integer, intent(IN)    ::  ictr, nnod, inod
      integer  ::  j_2, j_3

      j_2 = eigen_loop_start( ictr, nnod, inod )
      j_3 = eigen_loop_end  ( ictr, nnod, inod )
      if ( j_2 == j_3 ) then
         ret = j_2
      else
         ret = -1
      end if

      return
      end function  eigen_owner_index
!--   
!--   
      subroutine  eigen_vector_zeropad_x( v_x, L )
      use eigen_devel
      implicit NONE

      real(8), intent(INOUT) :: v_x(*)
      integer, intent(IN)    :: L
      integer                :: j_3, n

      j_3 = eigen_loop_end     (L, x_nnod, x_inod)
      n   = eigen_translate_g2l(L, x_nnod, x_inod)
      if ( j_3 < n ) then
         v_x(j_3+1:n) = ZERO    ! in case
      end if

      end subroutine  eigen_vector_zeropad_x
!--   
!--   
      subroutine  eigen_vector_zeropad_y( v_y, L )
      use eigen_devel
      implicit NONE

      real(8), intent(INOUT) :: v_y(*)
      integer, intent(IN)    :: L
      integer                :: j_3, n

      j_3 = eigen_loop_end     (L, y_nnod, y_inod)
      n   = eigen_translate_g2l(L, y_nnod, y_inod)
      if ( j_3 < n ) then
         v_y(j_3+1:n) = ZERO    ! in case
      end if

      end subroutine  eigen_vector_zeropad_y
!--   
!--   
      subroutine  eigen_vector_zeropad_x_lmask( v_x, L, is, ie )
      use eigen_devel
      implicit NONE

      real(8), intent(INOUT) :: v_x(*)
      integer, intent(IN)    :: L, is, ie
      integer                :: j_3, n

      if ( ie < is ) return
      j_3 = eigen_loop_end     (L, x_nnod, x_inod)
      n   = eigen_translate_g2l(L, x_nnod, x_inod)
      if ( j_3 < n ) then
         if ( j_3+1 < is ) then
            v_x(is:ie) = ZERO    ! in case
         else
         if ( j_3+1 <= ie ) then
         v_x(j_3+1:n) = ZERO    ! in case
         end if
         end if
      end if

      end subroutine  eigen_vector_zeropad_x_lmask
!--   
!--   
      subroutine  eigen_vector_zeropad_y_lmask( v_y, L, is, ie )
      use eigen_devel
      implicit NONE

      real(8), intent(INOUT) :: v_y(*)
      integer, intent(IN)    :: L, is, ie
      integer                :: j_3, n

      if ( ie < is ) return
      j_3 = eigen_loop_end     (L, y_nnod, y_inod)
      n   = eigen_translate_g2l(L, y_nnod, y_inod)
      if ( j_3 < n ) then
         if ( j_3+1 < is ) then
            v_y(is:ie) = ZERO    ! in case
         else
         if ( j_3+1 <= ie ) then
         v_y(j_3+1:n) = ZERO    ! in case
         end if
         end if
      end if

      end subroutine  eigen_vector_zeropad_y_lmask
!--   
!     
!     NOTICE : FOLLOWING ROUTINES ARE BASED ON AND LIMITTED FOR
!     THE MICROPROCESSORS, WHICH ADOPT THE IEEE754 FLOATING POINT
!     ARITHMETIC FORMAT AND OPERATIONS
!     
      real(8) function  machine_epsilon( )
     &     result( r )
      implicit NONE
      real(8) :: eps
      data eps /z'3CB0000000000000'/

      r = eps

      return
      end function  machine_epsilon
!--   
      real(8) function  get_constant_eps( )
     &     result( r )

      r = machine_epsilon( )

      return
      end function  get_constant_eps
!--   
      real(8) function  get_constant_pai( )
     &     result( r )
      implicit NONE
      real(8) :: const_pai
      data const_pai /z'400921FB54442D18'/

      r = const_pai

      return
      end function  get_constant_pai
!--   
      real(8) function  get_constant_2pai( )
     &     result( r )
      implicit NONE
      real(8) :: const_2pai
      data const_2pai /z'401921FB54442D18'/

      r = const_2pai

      return
      end function  get_constant_2pai
!--   
      real(8) function  get_constant_pai2( )
     &     result( r )
      implicit NONE
      real(8) :: const_pai2
      data const_pai2 /z'3ff921FB54442D18'/

      r = const_pai2

      return
      end function  get_constant_pai2
!--   
      end module  eigen_libs

