      subroutine  eigen_prd_ux(
     &     a, nm,
     &     u_x, u_y, nv,
     &     u_t, v_t, i, c, e, ne)
!$    use OMP_LIB
      use MPI
      use eigen_devel
      use eigen_libs
      use eigen_house, only : tsave
      implicit NONE

      integer, intent(IN)    :: nm, nv, ne, i
      real(8), intent(INOUT) :: a(1:nm,*)
      real(8), intent(OUT)   :: u_x(1:nv,*), u_y(1:nv,*)
      real(8), intent(OUT)   :: u_t(*), v_t(*)

      real(8), intent(OUT)   :: c(MBAND,MBAND), e(1:ne,*)

      real(8)                :: sgm(MBAND), bet(MBAND)
      real(8)                :: a_n, anorm2

      integer                :: x_owner_nod, x_pos
      integer                :: y_owner_nod, y_pos
      integer                :: i_1, i_2, i_3, i_4, i_5
      integer                :: j_1, j_2, j_3, j_4, j_5
      integer                :: k_1, k_2, k_3, k_4, k_5
      integer                :: jj_1, jj_2, jj_3
      integer                :: L, k
      integer                :: itr
      logical                :: mask(MBAND)

      real(8)                :: t, s, f
      real(8)                :: t11, t12, t22
      real(8)                :: s11, s12, s22
      real(8)                :: r12, rr(MBAND)
      real(8)                :: a0, a1, a2
!     real(8)                :: eps, tol

!     integer, parameter     :: VLEN = 2048
      integer, parameter     :: VLEN = 512

      real(8) :: d1,d2


      L  = i-MBAND


!$OMP MASTER

      y_owner_nod = eigen_owner_node   ( L+1, y_nnod, y_inod )
      x_pos       = eigen_translate_g2l( L+1, x_nnod, x_inod )
      call bcastw_dbl( a(1, 1), x_pos, y_owner_nod,
     &     nm, 2, u_t, 1, y_COMM_WORLD )


!     eps = machine_epsilon()

      j_2 = eigen_loop_start( 1, x_nnod, x_inod )
      j_3 = eigen_loop_end  ( L, x_nnod, x_inod )
      do j_1=j_2,j_3
         u_x(j_1, 2) = a(j_1, 2)
         u_x(j_1, 1) = a(j_1, 1)
      end do                    ! j_1

      r12 = ZERO
      do itr=1,2

                                ! compute a Gram matrix
         t11 = ZERO
         t12 = ZERO
         t22 = ZERO
         j_2 = eigen_loop_start( 1, x_nnod, x_inod )
         j_3 = eigen_loop_end  ( L, x_nnod, x_inod )
         do j_1=j_2,j_3
            t = u_x(j_1, 2)
            s = u_x(j_1, 1)
            t11 = t11 + t * t
            t12 = t12 + s * t
            t22 = t22 + s * s
         end do                 ! j_1

         u_t(1) = t11
         u_t(2) = t12
         u_t(3) = t22
         j_1 = eigen_owner_index( L-1, x_nnod, x_inod )
         j_2 = eigen_owner_index( L,   x_nnod, x_inod )
         j_3 = eigen_owner_index( L+1, x_nnod, x_inod )
         if ( j_1 > 0 ) then
            u_t(4) = u_x(j_1, 1)
         else
            u_t(4) = ZERO
         end if
         if ( j_2 > 0 ) then
            u_t(5) = u_x(j_2, 1)
         else
            u_t(5) = ZERO
         end if
         if ( itr == 1 ) then
            if ( j_1 > 0 ) then
               u_t(6) = u_x(j_1, 2)
            else
               u_t(6) = ZERO
            end if
            if ( j_2 > 0 ) then
               u_t(7) = u_x(j_2, 2)
            else
               u_t(7) = ZERO
            end if
            if ( j_3 > 0 ) then
               u_t(8) = a(j_3, 2)
            else
               u_t(8) = ZERO
            end if
            k = 8
         else
            k = 5
         end if
         call reduce_dbl( u_t, v_t, k, 3, x_COMM_WORLD )
         t11 = u_t(1)
         t12 = u_t(2)
         t22 = u_t(3)

                                !
                                ! do modified Cholesky factorization
                                ! [1    ][c1  ][1 s12] [c1 0     ][1 s12
                                !]
                                ! [s12 1][  c2][    1]=[s12*c1 c2][    1
                                !]
                                !                      [c1              
                                !    ]
                                !                     =[c1*s12 c1*s12*s1
                                !2+c2]
                                ! if g11 /= 0
                                ! -> c1=g11, s12=g12/c1, c2=g22-c1*s12**
                                !2
                                ! if g11 == 0
                                ! -> c1=0, s12=0, c2=g22
         if ( t11 == ZERO ) then
            mask(2) = .FALSE.
            s11 = ZERO
            s12 = ZERO
            s22 = t22
         else
            mask(2) = .TRUE.
            s11 = t11
            s12 = t12 / t11
            s22 = t22 - s11 * s12**2
         end if
         if ( s22 == ZERO ) then
            mask(1) = .FALSE.
         else
            mask(1) = .TRUE.
         endif

                                !
                                ! do multiply R^{-1} from right and comp
                                !ute Q
                                !
         j_2 = eigen_loop_start( 1, x_nnod, x_inod )
         j_3 = eigen_loop_end  ( L, x_nnod, x_inod )
         if ( mask(2) ) then
            if ( mask(1) ) then
               u_x(j_2:j_3,1) = u_x(j_2:j_3,1) - s12 * u_x(j_2:j_3,2)
               u_t(4) = u_t(4) - s12 * u_t(6)
               u_t(5) = u_t(5) - s12 * u_t(7)
            end if
         else
            u_x(j_2:j_3,2) = ZERO
            u_t(6) = ZERO
            u_t(7) = ZERO
         endif
         if ( mask(1) ) then
            continue
         else
            u_x(j_2:j_3,1) = ZERO
            u_t(4) = ZERO
            u_t(5) = ZERO
         end if

                                !
                                ! R_updated <= R_new * R_old
                                !  [1   s12][1 r12] = [1 r12+s12]
                                !  [    1  ][  1  ]   [  1      ]
                                !
         r12 = r12 + s12

         if ( itr == 1 ) then

                                !
                                ! if the condition number is good, skip 
                                !second QR step.
                                !
            s = MAX(s22, s11)
            t = MIN(s22, s11)
            if ( t * (2**20) > s ) then
               exit
!     !     else
!     !        if ( TRD_inod == 1 ) then
!     !           print*,"Do the second modified CholeskyQR", i
!     !        end if
            end if

         end if

      end do

      rr(1) = SQRT(s22)
      rr(2) = SQRT(s11)

      bet(1:2) = ONE
      sgm(1:2) = ZERO

                                !
                                ! Compute Householder reflectors
                                !
      if ( mask(2) ) then
         sgm(2) = -SIGN(rr(2), u_t(7))
         j_1 = eigen_owner_index( L, x_nnod, x_inod )
         if ( j_1 > 0 ) then
            u_x(j_1, 2) = u_x(j_1, 2) - sgm(2)
         end if
         u_t(7) =  u_t(7) - sgm(2)
         bet(2) = -u_t(7) * sgm(2)
         if ( mask(1) ) then
            s = sgm(2) * u_t(5) / bet(2)
            j_2 = eigen_loop_start( 1  , x_nnod, x_inod )
            j_3 = eigen_loop_end  ( L-1, x_nnod, x_inod )
            if ( j_2 <= j_3 ) then
               u_x(j_2:j_3, 1) = u_x(j_2:j_3, 1) + s * u_x(j_2:j_3, 2)
            end if
            u_t(4) = u_t(4) + s * u_t(6)
         end if
      end if
      if ( mask(1) ) then
         sgm(1) = -SIGN(rr(1), u_t(4))
         j_1 = eigen_owner_index( L-1, x_nnod, x_inod )
         if ( j_1 > 0 ) then
            u_x(j_1, 1) = u_x(j_1, 1) - sgm(1)
         end if
         u_t(4) =  u_t(4) - sgm(1)
         bet(1) = -u_t(4) * sgm(1)
      end if
      if ( mask(2) ) then
         j_1 = eigen_owner_index( L, x_nnod, x_inod )
         if ( j_1 > 0 ) then
            u_x(j_1, 1) = ZERO
         end if
         j_1 = eigen_owner_index( L+1, x_nnod, x_inod )
         if ( j_1 > 0 ) then
            u_x(j_1, 2) = ZERO
            a  (j_1, 2) = ZERO
         end if
      end if


                                ! [sgm(1)      0][1   0]
                                ! [ 0     sgm(2)][r12 1]
      e(i-1, 1) = sgm(2) * r12
      e(i  , 1) = u_t(8)
      e(i-1, 2) = sgm(1)
      e(i  , 2) = sgm(2)

      c(1,1) = ONE / bet(1)
      c(1,2) = ZERO
      c(2,1) = ZERO
      c(2,2) = ONE / bet(2)


      j_2 = eigen_loop_start( 1, x_nnod, x_inod )
      j_3 = eigen_loop_end  ( L, x_nnod, x_inod )
      if ( j_2 <= j_3 ) then
         a(j_2:j_3, 2) = u_x(j_2:j_3, 2)
         a(j_2:j_3, 1) = u_x(j_2:j_3, 1)
      end if


      x_pos = eigen_translate_g2l( L+1, x_nnod, x_inod )
      call datacast_dbl2( u_y(1, 1), u_y(1, 2),
     &     u_x(1, 1), u_x(1, 2), u_t, v_t, x_pos, 1 )

      call eigen_vector_zeropad_x( u_x(1, 1), L+1 )
      call eigen_vector_zeropad_x( u_x(1, 2), L+1 )
      call eigen_vector_zeropad_y( u_y(1, 1), L+1 )
      call eigen_vector_zeropad_y( u_y(1, 2), L+1 )

!$OMP END MASTER


      return
      end subroutine  eigen_prd_ux

