      subroutine  eigen_trd_u(
     &     d_out,
     &     u_x, u_y, nv,
     &     u_t, v_t, i, flag_concat, flag_sample, v_x, v_y)
!$    use OMP_LIB
      use MPI
      use eigen_devel
      use eigen_libs
      implicit NONE

      integer, intent(IN)    ::  nv, i
      real(8), intent(OUT)   ::  d_out
      real(8), intent(INOUT) ::  u_x(1:nv), u_y(1:nv)
      real(8), intent(INOUT) ::  u_t(*), v_t(*)
      logical, intent(IN)    ::  flag_concat, flag_sample
      real(8), intent(IN)    ::  v_x(*)
      real(8), intent(OUT)   ::  v_y(*)

      integer                ::  x_pos, y_pos
      integer                ::  j_1, j_2, j_3
      integer                ::  L


      L = i-1

!
! Pick up the diagonal element from the target vector in panel (u_x)
! into d_out, then nullfy the element on the panel.
!
      j_1 = eigen_owner_index(i, x_nnod, x_inod)
      if ( j_1 > 0 ) then
         d_out = u_x(j_1)
         u_x(j_1) = ZERO
      else
         d_out = ZERO
      end if


      if ( flag_overlap .OR. flag_oncache ) then

         x_pos = eigen_translate_g2l(L, x_nnod, x_inod)
         call datacast_dbl( u_y(1), u_x(1), u_t(1), v_t(1), x_pos, 1 )

         call eigen_vector_zeropad_x( u_x(1), L )
         call eigen_vector_zeropad_y( u_y(1), L )

      else

         if ( flag_concat ) then

            x_pos = eigen_translate_g2l(i, x_nnod, x_inod)

         if ( .NOT. flag_oncache ) then
         if ( flag_sample ) then
            timer_t1 = eigen_get_wtime()
         end if
         end if
            call datacast_dbl2( u_y(1), v_y(1), u_x(1), v_x(1),
     &                          u_t(1), v_t(1), x_pos, 1 )
         if ( .NOT. flag_oncache ) then
         if ( flag_sample ) then
            timer_t1 = eigen_get_wtime() - timer_t1 - Bcast_Overhead_x
            timer_t1 = timer_t1/2 + Bcast_cont_Overhead_x
         end if
         end if

         else

            x_pos = eigen_translate_g2l(L, x_nnod, x_inod)
            call datacast_dbl( u_y(1), u_x(1),
     &                         u_t(1), v_t(1), x_pos, 1 )

         end if

            call eigen_vector_zeropad_x( u_x(1), L )
            call eigen_vector_zeropad_y( u_y(1), L )
         if ( flag_concat ) then
            call eigen_vector_zeropad_y( v_y(1), i )
         end if

      end if

!$OMP BARRIER

      if ( .NOT. flag_oncache ) then
      if ( flag_overlap ) then

         if ( flag_concat ) then

            x_pos = eigen_translate_g2l(i, x_nnod, x_inod)

         if ( flag_sample ) then
            timer_t1 = eigen_get_wtime()
         end if
            call datacast_dbl( v_y(1), v_x(1),
     &                         v_y(1-nv), v_x(1-nv), x_pos, 1 )
         if ( flag_sample ) then
            timer_t1 = eigen_get_wtime() - timer_t1
         end if

            call eigen_vector_zeropad_y( v_y(1), i )

         end if

      end if
      end if


      return
      end subroutine  eigen_trd_u

