      subroutine  eigen_trd_init(a, nm, n,
     &     d_out, e_out,
     &     u_t, v_t, nv)
!$    use omp_lib
      use eigen_devel
      use eigen_libs
      use eigen_house
      implicit NONE

      integer, intent(IN)    ::  nm, n, nv
      real(8), intent(INOUT) ::  a(1:nm, *)
      real(8), intent(OUT)   ::  d_out(*)
      real(8), intent(OUT)   ::  e_out(*)
      real(8), intent(IN)    ::  u_t(*), v_t(*)

      integer                ::  i_1, i_2, i_3, i_4
      integer                ::  j_1, j_2, j_3, j_4
      integer                ::  k_1, k_2, k_3, k_4
      integer                ::  i, j

      integer                ::  local_size, local_rank

      include 'CSTAB.h'


      local_rank = 0
      local_size = 1
!$    local_rank = omp_get_thread_num()
!$    local_size = omp_get_num_threads()

      if ( local_size > 1 .AND. local_rank == 1 ) then

         allocate( u0_z(nv*local_size+n_columns) )
         allocate( v0_z(nv*local_size+n_columns) )
         call CSTAB_adjust_base( u0_z(1), u_t(1), offset1 )
         call CSTAB_adjust_base( v0_z(1), v_t(1), offset2 )

      end if

      i_2 = eigen_loop_start(1, y_nnod, y_inod)
      i_3 = eigen_loop_end  (n, y_nnod, y_inod)

!$OMP DO SCHEDULE(DYNAMIC)
      do  i_1 = i_2, i_3
         j   = eigen_translate_l2g(i_1, y_nnod, y_inod)
         j_2 = eigen_loop_start   (j+1, x_nnod, x_inod)
         if ( j <= n ) then
            a(j_2:nm, i_1) = ZERO
         else
            a(1:nm, i_1) = ZERO
         end if
      end do                    ! i_1
!$OMP END DO

!$OMP MASTER

      flag_overlap = .TRUE.
      timer_t1 = ZERO
      timer_t2 = ONE

!$OMP END MASTER


      return
      end subroutine  eigen_trd_init


      subroutine  eigen_trd_final(a, nm, n, d_out, e_out, u_t)
!$    use omp_lib
      use eigen_devel
      use eigen_libs
      use eigen_house
      implicit NONE

      integer, intent(IN)    ::  nm, n
      real(8), intent(INOUT) ::  a(1:nm, *)
      real(8), intent(OUT)   ::  d_out(*)
      real(8), intent(OUT)   ::  e_out(*)
      real(8), intent(OUT)   ::  u_t(*)

      integer                ::  x_pos, x_owner_nod
      integer                ::  y_pos, y_owner_nod
      integer                ::  i_1, i_2, i_3, i_4
      integer                ::  j_1, j_2, j_3, j_4
      integer                ::  k_1, k_2, k_3, k_4
      integer                ::  i, j, k, L
      real(8)                ::  t

      integer                ::  local_size, local_rank

      include 'CSTAB.h'


      local_rank = 0
      local_size = 1
!$    local_rank = omp_get_thread_num()
!$    local_size = omp_get_num_threads()

      if ( local_size > 1 .AND. local_rank == 1 ) then

         deallocate ( u0_z )
         deallocate ( v0_z )

      end if

      if ( n >= 2 .AND. local_rank == 0 ) then

         i = 2; L = i-1
         i_1 = eigen_owner_index (i, y_nnod, y_inod)
         j_1 = eigen_owner_index (L, x_nnod, x_inod)
         if ( i_1 > 0 .AND. j_1 > 0 ) then
            t           =   a(j_1, i_1)
            u_t(1)      = - t
            a(j_1, i_1) = 2*t
         else
            u_t(1)      = ZERO
         end if

         do i=1,2
            i_1 = eigen_owner_index (i, y_nnod, y_inod)
            j_1 = eigen_owner_index (i, x_nnod, x_inod)
            if ( i_1 > 0 .AND. j_1 > 0 ) then
               u_t(1+i) = a(j_1, i_1)
            else
               u_t(1+i) = ZERO
            end if
         enddo

         call reduce_dbl ( u_t(1), u_t(4), 3, 2, x_COMM_WORLD )
         call reduce_dbl ( u_t(1), u_t(4), 3, 2, y_COMM_WORLD )

         e_out(1) = ZERO
         e_out(2) = u_t(1)
         d_out(1) = u_t(2)
         d_out(2) = u_t(3)

      end if

      
      return
      end subroutine  eigen_trd_final

